/*
 * This code is part of the strategy game operational-space.
 * operational-space comes with ABSOLUTELY NO WARRANTY and is licensed under GPL-2.0.
 *    Copyright (C) 2024  VegOwOtenks, Sleppo04
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#include "dynamicarray.h"

int DynamicArray_Create(DynamicArray* target, size_t object_size, size_t initial_capacity, allocator_t* allocator)
{
	if (target == NULL) {
		return EDESTADDRREQ;
	}
	if (object_size == 0) {
		return ERANGE;
	}
	if (initial_capacity == 0) {
		return EINVAL;
	}

	target->memory = Allocator_Allocate(allocator, object_size * initial_capacity);
	if (target->memory == NULL) {
		return ENOMEM;
	}
	target->object_size = object_size;
	target->capacity    = initial_capacity;
	target->allocator   = allocator;
	target->reserved    = 0;

	return EXIT_SUCCESS;
}

int DynamicArray_Clone(DynamicArray* target, DynamicArray* original)
{
	if (target == NULL) {
		return EDESTADDRREQ;
	}
	if (original == NULL) {
		return EINVAL;
	}

	*target = *original;
	target->memory = Allocator_AllocateArray(target->allocator, target->object_size, target->capacity);
	if (target->memory == NULL) {
		return ENOMEM;
	}

	memcpy(target->memory, original->memory, original->object_size * original->capacity);

	return EXIT_SUCCESS;
}

int DynamicArray_DeepClone(DynamicArray* target, DynamicArray* original, DynamicArrayCloneFunction clone, void* xarg)
{
	if (target == NULL) {
		return EDESTADDRREQ;
	}
	if (original == NULL) {
		return EINVAL;
	}
	if (clone == NULL) {
		return EINVAL;
	}

	*target = *original;
	target->memory = Allocator_AllocateArray(target->allocator, target->object_size, target->capacity);
	if (target->memory == NULL) {
		return ENOMEM;
	}

	DYNAMICARRAY_FOREACH(*original, i) {
		void* clone_current    = DynamicArray_GetPointer(target, i);
		void* original_current = DynamicArray_GetPointer(target, i);
		int clone_code = clone(clone_current, original_current, xarg);
		if (clone_code) {
			return ECANCELED;
		}
	}

	return EXIT_SUCCESS;
}

int DynamicArray_ResizeObjects(DynamicArray* array, size_t new_object_size)
{
    if (array == NULL) {
        return EDESTADDRREQ;
    }
    if (new_object_size == 0) {
        return ERANGE;
    }

    size_t old_object_size = array->object_size;
    size_t old_size = old_object_size * array->capacity;
    size_t new_size = new_object_size * array->capacity;
    void* new_array = Allocator_Reallocate(array->allocator, array->memory, old_size, new_size);
    if (new_array == NULL) {
        return ENOMEM;
    }
    array->memory = new_array;

    if (array->reserved != 0) {
        // move old stuff
        void* current = advancep(array->memory, old_object_size * (array->reserved - 1));
        size_t index = array->reserved - 1;
        while (index != 0) {
            void* location    = current;
            void* destination = DynamicArray_GetPointer(array, index);
            memmove(destination, location, old_object_size);

            current = rewindp(current, old_object_size);
            index--;
        }
    }

    array->object_size = new_object_size;

    return EXIT_SUCCESS;
}

int DynamicArray_Resize(DynamicArray* array, size_t new_capacity)
{
	if (array == NULL) {
		return EDESTADDRREQ;
	}
	if (new_capacity == 0) {
		return ERANGE;
	}

	size_t old_size = array->object_size * array->capacity;
	size_t new_size = array->object_size * new_capacity;
	void* new_array = Allocator_Reallocate(array->allocator, array->memory, old_size, new_size);
	if (new_array == NULL) {
		return ENOMEM;
	}
	array->memory = new_array;
	array->capacity = new_capacity;

	return EXIT_SUCCESS;
}

int DynamicArray_Append(DynamicArray* array, void* object)
{
	if (array == NULL) {
		return EDESTADDRREQ;
	}
	if (object == NULL) {
		return EINVAL;
	}

	if (array->capacity == array->reserved) {
		if (DynamicArray_Resize(array, array->capacity * 2)) {
			return ENOMEM;
		}
	}

	void* destination = advancep(array->memory, array->reserved * array->object_size);
	memcpy(destination, object, array->object_size);
	array->reserved++;

	return EXIT_SUCCESS;
}

int DynamicArray_AppendEmpty(DynamicArray* array, void** pointer)
{
	if (array == NULL) {
		return EDESTADDRREQ;
	}

	if (array->capacity == array->reserved) {
		if (DynamicArray_Resize(array, array->capacity * 2)) {
			*pointer = NULL;
			return ENOMEM;
		}
	}

	array->reserved++;

	*pointer = DynamicArray_GetPointer(array, array->reserved - 1);

	return EXIT_SUCCESS;
}

int DynamicArray_Remove(DynamicArray* array, size_t index)
{
	if (array == NULL) {
		return EDESTADDRREQ;
	}
	if (index >= array->reserved) {
		return EBOUNDS;
	}

	if (index != array->reserved - 1) {
		void* to    = advancep(array->memory, array->object_size * index);
		void* from  = advancep(to, array->object_size);
		size_t size = array->object_size * (array->reserved - index);
		memmove(to, from, size);
	}
	array->reserved--;

	return EXIT_SUCCESS;
}

int DynamicArray_RemoveFast(DynamicArray* array, size_t index)
{
	if (array == NULL) {
		return EDESTADDRREQ;
	}
	if (index >= array->reserved) {
		return EBOUNDS;
	}

	if (array->capacity > 1) {
		void* to   = DynamicArray_GetPointer(array, index);
		void* from = DynamicArray_GetPointer(array, array->reserved - 1);
		memcpy(to, from, array->object_size);
	}
	array->reserved--;

	return EXIT_SUCCESS;
}

size_t DynamicArray_FindFunction(DynamicArray* array, DynamicArrayFindFunction function, void* xarg)
{
	if (array == NULL) return SIZE_MAX;
	if (function == NULL) return SIZE_MAX;

	DYNAMICARRAY_FOREACH(*array, i) {
		void* current = DynamicArray_GetPointer(array, i);
		if (function(current, xarg)) {
			return i;
		}
	}

	return SIZE_MAX;
}

size_t DynamicArray_FindFunctionLinear(DynamicArray* array, DynamicArrayLinearFindFunction function, void* xarg)
{
	if (array == NULL) return SIZE_MAX;
	if (function == NULL) return SIZE_MAX;

	size_t bot  = 0;
	size_t top  = array->reserved;
	size_t mid  = bot + (top - bot) / 2;
	int    eval = -1;

	while (eval != 0 && mid != top && mid < array->reserved) {
		void* current = DynamicArray_GetPointer(array, mid);
		eval = function(current, xarg);

		if (eval == 0) {
			bot = top = mid;
		} else if (eval < 0) {
			top = mid - 1;
		} else {
			bot = mid + 1;
		}
		// TODO: Is there a bug with overlapping top and bot here?

		mid  = bot + (top - bot) / 2;
	}

	if (mid == top && mid < array->reserved) {
		void* current = DynamicArray_GetPointer(array, mid);
		eval = function(current, xarg);
	}

	if (eval != 0) {
		return SIZE_MAX;
	}

	return mid;
}

void* DynamicArray_GetPointer(DynamicArray* array, size_t index)
{
	if (array == NULL) {
		return NULL;
	}
	if (index >= array->reserved) {
		return NULL;
	}

	size_t offset = index * array->object_size;

	return advancep(array->memory, offset);
}

size_t DynamicArray_GetLength(DynamicArray* array)
{
	return array->reserved;
}

void DynamicArray_Destroy(DynamicArray* array)
{
	if (array == NULL) {
		return;
	}

	Allocator_Free(array->allocator, array->memory, array->object_size * array->capacity);
	memset(array, 0, sizeof(*array));

	return;
}
